==> hperf-6.8.8.tar.gz: hperf/main.c <==
/*
    This file is part of HPerf.
    Copyright (C) 2020-2023  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "message.h"
#include "pipe.h"
#include "prog.h"
#include "trace.h"
#include "meta.h"
#include "dump.h"
#include "output.h"
#include "main.h"

// ************************************************************************
// 
// ************************************************************************
struct param_info {
	char *flag;
	char *arg_name;
	char *description;
	char *def_val;
};

enum param_id {
	PARAM_INPUT,
	PARAM_OUTPUT,
	PARAM_SAMPLE_THRESHOLD,
	PARAM_HOTSPOT_THRESHOLD,
	PARAM_HOTSPOT_CONTEXT,
	PARAM_DUMP_CONTEXT,
	PARAM_CSS_BASE,
	PARAM_CSS_ADD,
	PARAM_THEME,
	PARAM_VERBOSITY,
	NPARAMS,
};

static struct param_info param[NPARAMS] = {
{ "-i", "file", "input file, produced by perf-record", "perf.data" },
{ "-o", "file", "output file", "report.html" },
{ "-s", "count[%]", "minimum number of samples per insn", "1" },
{ "-t", "count[%]", "minimum total number of samples per hotspot", "2" },
{ "-c", "n", "merge hotspots separated by up to n insn", "5" },
{ "-d", "n", "output n insn before and after hotspots", "100" },
{ "-S", "file", "replacement css file path", NULL },
{ "-A", "file", "additional css file path", NULL },
{ "-T", "theme", "initial theme: 'dark' or 'light'", "light" },
{ "-v", "level", "verbosity level", "1" },
};

// ************************************************************************
// 
// ************************************************************************
static void help_message(void)
{
	MESSAGE("Usage: hperf [options]\n"
		"\n"
		"Options:\n"
		"\n");
	
	for (int p = 0; p < NPARAMS; p++) {
		MESSAGE("  %-4s %-10s   %s (default: %s)\n",
			param[p].flag, param[p].arg_name,
			param[p].description,
			(param[p].def_val) ? param[p].def_val : "none");
	}
	
	MESSAGE("\n");
}

// ************************************************************************
// 
// ************************************************************************
#ifndef HPERF_VERSION
#define HPERF_VERSION "0.0.0"
#endif

static int version_parse(char *s, int *v, int n)
{
	for (int i = 0; i < n; i++) {
		char *e;
		int u = strtol(s, &e, 10);
		
		if (e == s)
			return -1;
		
		if (v)
			v[i] = u;
		
		if (*e == 0)
			return 0;
		
		if (*e != '.')
			return -1;
		
		s = e + 1;
	}
	
	return -1;
}

static int perf_version(int *v, int n)
{
	char buff[256];
	char *argv[3] = { "perf", "--version", NULL };
	
	struct piped_input pi;
	
	if (pipe_in(&pi, argv, NULL))
		return -1;
	
	size_t rn = fread(buff, 1, sizeof(buff) - 1, pi.file);
	int error = ferror(pi.file);
	int pr = pipe_wait(&pi);
	
	if (error)
		ERROR("perf --version: %s\n", strerror(errno));
	
	if (pr) 
		ERROR("perf --version: failed with status %d\n", pr);
	
	if ((error) || (pr))
		return -1;
	
	buff[rn] = 0;
	
	if (strncmp(buff, "perf version ", 13) != 0) {
		ERROR("perf: '%s'\n", buff);
		return -1;
	}
	
	return version_parse(buff + 13, v, n);
}

static char hperf_version_info[256] = "";

static int version_get(char **version_info_r, int *supported_r)
{
	int p[3] = { 0, 0, 0 };
	int h[3] = { 0, 0, 0 };
	
	int r = 0;
	
	r |= perf_version(p, 3);
	r |= version_parse(HPERF_VERSION, h, 3);
	
	int supported = (p[0] < h[0]) || ((p[0] == h[0]) && (p[1] <= h[1]));
	
	snprintf(hperf_version_info, sizeof(hperf_version_info),
		"  perf %d.%d.%d\n hperf %d.%d.%d\n",
		p[0], p[1], p[2], h[0], h[1], h[2]);
	
	if (supported_r)
		*supported_r = supported;
	
	if (version_info_r)
		*version_info_r = hperf_version_info;
	
	return r;
}

// ************************************************************************
// 
// ************************************************************************
static int arg_parse(char **val, char *flag, char *arg)
{
	int p = 0;
	
	while (p < NPARAMS) {
		if (strcmp(flag, param[p].flag) == 0)
			break;
		p++;
	}
	
	if (p >= NPARAMS) {
		ERROR("Unknown flag '%s'.\n", flag);
		return -1;
	}
	
	if (arg == NULL) {
		ERROR("Value expected after '%s'.\n", flag);
		return -1;
	}
	
	if (val[p] != NULL) {
		ERROR("Flag '%s' specified multiple times.\n", flag);
		return -1;
	}
	
	val[p] = arg;
	
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
static int pcv(char *str, int *dst)
{
	char *e;
	
	long r = strtol(str, &e, 0);
	
	if ((e == str) || (*e != 0)) {
		ERROR("%s: could not parse integer\n", str);
		return -1;
	}
	
	if (dst)
		*dst = r;
	
	return 0;
}

static int pci(char *str, uint64_t *dst)
{
	char *e;
	
	uint64_t r = strtol(str, &e, 0);
	
	if ((e == str) || (*e != 0)) {
		ERROR("%s: could not parse insn count\n", str);
		return -1;
	}
	
	if (dst)
		*dst = r;
	
	return 0;
}

static int pcs(char *str, uint64_t *dst, uint64_t total)
{
	char *e;
	
	uint64_t r = strtol(str, &e, 0);
	
	if ((e == str) || (*e != 0)) {
		double f = strtod(str, &e);
		
		if ((e == str) || (e[0] != '%') || (e[1] != 0)) {
			ERROR("%s: could not parse sample count\n", str);
			return -1;
		}
		
		r = f * 0.01 * total;
		
		MESSAGE("  note: %s == %ld samples\n", str, r);
	}
	
	if (dst)
		*dst = r;
	
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
static int run(char **val, char *version_info)
{
	if (pcv(val[PARAM_VERBOSITY], &verbosity))
		return -1;
	
	struct prog prog;
	struct meta meta;
	
	prog_init(&prog);
	meta_init(&meta);
	
	int r = trace_load(&prog, val[PARAM_INPUT]);
	
	if (r)
		goto clear;
	
	if (prog.samples == 0) {
		ERROR("No samples read.\n");
		r = -1;
		goto clear;
	}
	
	r |= pcs(val[PARAM_SAMPLE_THRESHOLD],
		&meta.sample_threshold_hits, prog.samples);
	r |= pcs(val[PARAM_HOTSPOT_THRESHOLD],
		&meta.hotspot_threshold_hits, prog.samples);
	r |= pci(val[PARAM_HOTSPOT_CONTEXT],
		&meta.hotspot_context_insn);
	r |= pci(val[PARAM_DUMP_CONTEXT],
		&meta.dump_context_insn);
	
	if (r)
		goto clear;
	
	r |= meta_run(&meta, &prog);
	
	if (r)
		goto clear;
	
	//dump(&meta, &prog);
	
	r = output(&meta, &prog, version_info, val[PARAM_OUTPUT],
		val[PARAM_CSS_BASE], val[PARAM_CSS_ADD], val[PARAM_THEME]);

clear:	
	prog_clear(&prog);
	meta_clear(&meta);

	return r;
}

// ************************************************************************
// 
// ************************************************************************
int main(int argc, char **argv)
{
	int show_version = 0;
	char *val[NPARAMS];
	
	for (int p = 0; p < NPARAMS; p++)
		val[p] = NULL;
	
	for (int t = 1; t < argc; t += 2) {
		if ((strcmp(argv[t], "-h") == 0)
		||  (strcmp(argv[t], "--help") == 0)) {
			help_message();
			return 0;
		}
		
		// Yes, "-v" and "-v 2" have different meanings :-/
		if (((strcmp(argv[t], "-v") == 0) && (t + 1 >= argc))
		||  (strcmp(argv[t], "--version") == 0)) {
			show_version = 1;
			continue;
		}
		
		if (arg_parse(val, argv[t], argv[t + 1])) {
			help_message();
			return 1;
		}
	}
	
	for (int p = 0; p < NPARAMS; p++) {
		if (val[p] == NULL)
			val[p] = param[p].def_val;
	}
	
	char *version_info;
	int supported;
	
	version_get(&version_info, &supported);
	
	if (show_version || !supported)
		MESSAGE("%s", version_info);
	
	if (!supported)
		MESSAGE("\nWarning: unsupported perf version\n\n");
	
	if (show_version)
		return 0;
	
	return run(val, version_info);
}

==> hperf-6.11.10.tar.gz: hperf/main.c <==
/*
    This file is part of HPerf.
    Copyright (C) 2020-2024  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "message.h"
#include "pipe.h"
#include "prog.h"
#include "trace.h"
#include "meta.h"
#include "dump.h"
#include "output.h"
#include "main.h"

// ************************************************************************
// 
// ************************************************************************
struct param_info {
	char *flag;
	char *arg_name;
	char *description;
	char *def_val;
};

enum param_id {
	PARAM_INPUT,
	PARAM_OUTPUT,
	PARAM_SAMPLE_THRESHOLD,
	PARAM_HOTSPOT_THRESHOLD,
	PARAM_HOTSPOT_CONTEXT,
	PARAM_DUMP_CONTEXT,
	PARAM_CSS_BASE,
	PARAM_CSS_ADD,
	PARAM_THEME,
	PARAM_VERBOSITY,
	NPARAMS,
};

static struct param_info param[NPARAMS] = {
{ "-i", "file", "input file, produced by perf-record", "perf.data" },
{ "-o", "file", "output file", "report.html" },
{ "-s", "count[%]", "minimum number of samples per insn", "1" },
{ "-t", "count[%]", "minimum total number of samples per hotspot", "2" },
{ "-c", "n", "merge hotspots separated by up to n insn", "5" },
{ "-d", "n", "output n insn before and after hotspots", "100" },
{ "-S", "file", "replacement css file path", NULL },
{ "-A", "file", "additional css file path", NULL },
{ "-T", "theme", "initial theme: 'dark' or 'light'", "light" },
{ "-v", "level", "verbosity level", "1" },
};

// ************************************************************************
// 
// ************************************************************************
static void help_message(void)
{
	MESSAGE("Usage: hperf [options]\n"
		"\n"
		"Options:\n"
		"\n");
	
	for (int p = 0; p < NPARAMS; p++) {
		MESSAGE("  %-4s %-10s   %s (default: %s)\n",
			param[p].flag, param[p].arg_name,
			param[p].description,
			(param[p].def_val) ? param[p].def_val : "none");
	}
	
	MESSAGE("\n");
}

// ************************************************************************
// 
// ************************************************************************
#ifndef HPERF_VERSION
#define HPERF_VERSION "0.0.0"
#endif

static int version_parse(char *s, int *v, int n)
{
	for (int i = 0; i < n; i++) {
		char *e;
		int u = strtol(s, &e, 10);
		
		if (e == s)
			return -1;
		
		if (v)
			v[i] = u;
		
		if (*e == 0)
			return 0;
		
		if (*e != '.')
			return -1;
		
		s = e + 1;
	}
	
	return -1;
}

static int perf_version(int *v, int n)
{
	char buff[256];
	char *argv[3] = { "perf", "--version", NULL };
	
	struct piped_input pi;
	
	if (pipe_in(&pi, argv, NULL))
		return -1;
	
	size_t rn = fread(buff, 1, sizeof(buff) - 1, pi.file);
	int error = ferror(pi.file);
	int pr = pipe_wait(&pi);
	
	if (error)
		ERROR("perf --version: %s\n", strerror(errno));
	
	if (pr) 
		ERROR("perf --version: failed with status %d\n", pr);
	
	if ((error) || (pr))
		return -1;
	
	buff[rn] = 0;
	
	if (strncmp(buff, "perf version ", 13) != 0) {
		ERROR("perf: '%s'\n", buff);
		return -1;
	}
	
	return version_parse(buff + 13, v, n);
}

static char hperf_version_info[256] = "";

static int version_get(char **version_info_r, int *supported_r)
{
	int p[3] = { 0, 0, 0 };
	int h[3] = { 0, 0, 0 };
	
	int r = 0;
	
	r |= perf_version(p, 3);
	r |= version_parse(HPERF_VERSION, h, 3);
	
	int supported = (p[0] < h[0]) || ((p[0] == h[0]) && (p[1] <= h[1]));
	
	snprintf(hperf_version_info, sizeof(hperf_version_info),
		"  perf %d.%d.%d\n hperf %d.%d.%d\n",
		p[0], p[1], p[2], h[0], h[1], h[2]);
	
	if (supported_r)
		*supported_r = supported;
	
	if (version_info_r)
		*version_info_r = hperf_version_info;
	
	return r;
}

// ************************************************************************
// 
// ************************************************************************
static int arg_parse(char **val, char *flag, char *arg)
{
	int p = 0;
	
	while (p < NPARAMS) {
		if (strcmp(flag, param[p].flag) == 0)
			break;
		p++;
	}
	
	if (p >= NPARAMS) {
		ERROR("Unknown flag '%s'.\n", flag);
		return -1;
	}
	
	if (arg == NULL) {
		ERROR("Value expected after '%s'.\n", flag);
		return -1;
	}
	
	if (val[p] != NULL) {
		ERROR("Flag '%s' specified multiple times.\n", flag);
		return -1;
	}
	
	val[p] = arg;
	
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
static int pcv(char *str, int *dst)
{
	char *e;
	
	long r = strtol(str, &e, 0);
	
	if ((e == str) || (*e != 0)) {
		ERROR("%s: could not parse integer\n", str);
		return -1;
	}
	
	if (dst)
		*dst = r;
	
	return 0;
}

static int pci(char *str, uint64_t *dst)
{
	char *e;
	
	uint64_t r = strtol(str, &e, 0);
	
	if ((e == str) || (*e != 0)) {
		ERROR("%s: could not parse insn count\n", str);
		return -1;
	}
	
	if (dst)
		*dst = r;
	
	return 0;
}

static int pcs(char *str, uint64_t *dst, uint64_t total)
{
	char *e;
	
	uint64_t r = strtol(str, &e, 0);
	
	if ((e == str) || (*e != 0)) {
		double f = strtod(str, &e);
		
		if ((e == str) || (e[0] != '%') || (e[1] != 0)) {
			ERROR("%s: could not parse sample count\n", str);
			return -1;
		}
		
		r = f * 0.01 * total;
		
		MESSAGE("  note: %s == %ld samples\n", str, r);
	}
	
	if (dst)
		*dst = r;
	
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
static int run(char **val, char *version_info)
{
	if (pcv(val[PARAM_VERBOSITY], &verbosity))
		return -1;
	
	struct prog prog;
	struct meta meta;
	
	prog_init(&prog);
	meta_init(&meta);
	
	int r = trace_load(&prog, val[PARAM_INPUT]);
	
	if (r)
		goto clear;
	
	if (prog.samples == 0) {
		ERROR("No samples read.\n");
		r = -1;
		goto clear;
	}
	
	r |= pcs(val[PARAM_SAMPLE_THRESHOLD],
		&meta.sample_threshold_hits, prog.samples);
	r |= pcs(val[PARAM_HOTSPOT_THRESHOLD],
		&meta.hotspot_threshold_hits, prog.samples);
	r |= pci(val[PARAM_HOTSPOT_CONTEXT],
		&meta.hotspot_context_insn);
	r |= pci(val[PARAM_DUMP_CONTEXT],
		&meta.dump_context_insn);
	
	if (r)
		goto clear;
	
	r |= meta_run(&meta, &prog);
	
	if (r)
		goto clear;
	
	//dump(&meta, &prog);
	
	r = output(&meta, &prog, version_info, val[PARAM_OUTPUT],
		val[PARAM_CSS_BASE], val[PARAM_CSS_ADD], val[PARAM_THEME]);

clear:	
	prog_clear(&prog);
	meta_clear(&meta);

	return r;
}

// ************************************************************************
// 
// ************************************************************************
int main(int argc, char **argv)
{
	int show_version = 0;
	char *val[NPARAMS];
	
	for (int p = 0; p < NPARAMS; p++)
		val[p] = NULL;
	
	for (int t = 1; t < argc; t += 2) {
		if ((strcmp(argv[t], "-h") == 0)
		||  (strcmp(argv[t], "--help") == 0)) {
			help_message();
			return 0;
		}
		
		// Yes, "-v" and "-v 2" have different meanings :-/
		if (((strcmp(argv[t], "-v") == 0) && (t + 1 >= argc))
		||  (strcmp(argv[t], "--version") == 0)) {
			show_version = 1;
			continue;
		}
		
		if (arg_parse(val, argv[t], argv[t + 1])) {
			help_message();
			return 1;
		}
	}
	
	for (int p = 0; p < NPARAMS; p++) {
		if (val[p] == NULL)
			val[p] = param[p].def_val;
	}
	
	char *version_info;
	int supported;
	
	version_get(&version_info, &supported);
	
	if (show_version || !supported)
		MESSAGE("%s", version_info);
	
	if (!supported)
		MESSAGE("\nWarning: perf version may not be supported\n\n");
	
	if (show_version)
		return 0;
	
	return run(val, version_info);
}
